# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

import json

NOTIFICATION = "NOTIFICATION"
CHAT = "CHAT"

NOTIFICATION_TYPES = [
    NOTIFICATION, CHAT
]


def send_notification(user, id, subject, body, created):
    """
    :param user:
    :param id:
    :param subject:
    :param body:
    :param created:
    :return:
    """
    # importing channels at module level would cause certain things to fail
    # when channels isn't installed, e.g. `manage.py prep` in `lino_book.projects.workflows`.
    from channels.layers import get_channel_layer
    from asgiref.sync import async_to_sync

    msg = dict(
        type=NOTIFICATION,
        subject=subject,
        id=id,
        body=body,
        created=created,
    )
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)(user.username,
                                            {"type": "send_notification",  # method name in consumer
                                             "text": json.dumps(msg)})  # data



# Copyright 2016-2018 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)
"""
"""
#
# import datetime
# from django.utils import translation
# from atelier.utils import i2t
from lino.api import dd, rt, _
#
# from django.conf import settings
# from django.utils.timezone import make_aware
from lino.modlib.chat.models import ChatGroup, ChatGroupMember


def objects():
    groups = ['General', 'Customers request']
    for group in groups:
        g = ChatGroup(title=group)
        g.save()
        yield g

        for u in rt.models.users.User.objects.order_by('username'):

            gm = ChatGroupMember(group=g, user=u)
            gm.save()
            yield gm

    # todo Create
    # todo spawn some random chats




# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from lino.api import dd, rt, _

from django import http
from django.conf import settings
from django.utils.translation import get_language
from lino.modlib.printing.mixins import Printable
from lino.modlib.printing.choicelists import BuildMethods


class Publishable(Printable):
    class Meta:
        abstract = True
        app_label = 'publisher'

    publisher_location = None
    publisher_template = "publisher/default.pub.html"

    def publisher_url(self):
        return "/{}/{}".format(self.publisher_location, self.pk)

    def get_publisher_response(self, ar):
        env = settings.SITE.plugins.jinja.renderer.jinja_env
        template = env.get_template(self.publisher_template)
        context = ar.get_printable_context(obj=self)
        # context = dict(obj=self, request=request, language=get_language())
        response = http.HttpResponse(
            template.render(**context),
            content_type='text/html;charset="utf-8"')
        return response

    @classmethod
    def get_dashboard_items(cls, user):
        return []



# -*- coding: UTF-8 -*-
# Copyright 2009-2019 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from builtins import object
from builtins import str

from django.conf import settings
from django.utils.encoding import force_text

from django.db import models
from django.utils.translation import ugettext_lazy as _

from django.apps import apps ; get_models = apps.get_models

from lino.api import dd, rt
from lino.core import actions
from lino.core.utils import full_model_name
from lino.core.roles import SiteStaff

from lino.modlib.printing.choicelists import BuildMethods
from lino.modlib.checkdata.choicelists import Checker


# import them here to have them on rt.models.system:
from .choicelists import YesNo, Genders, PeriodEvents
from .mixins import Lockable


class BuildSiteCache(dd.Action):
    label = _("Rebuild site cache")
    url_action_name = "buildjs"

    def run_from_ui(self, ar):
        settings.SITE.kernel.default_renderer.build_site_cache(True)
        return ar.success(
            """\
Seems that it worked. Refresh your browser.
<br>
Note that other users might experience side effects because
of the unexpected .js update, but there are no known problems so far.
Please report any anomalies.""",
            alert=_("Success"))


class SiteConfigManager(models.Manager):

    def get(self, *args, **kwargs):
        return settings.SITE.site_config



class SiteConfig(dd.Model):

    class Meta(object):
        abstract = dd.is_abstract_model(__name__, 'SiteConfig')
        verbose_name = _("Site configuration")

    objects = SiteConfigManager()
    real_objects = models.Manager()

    default_build_method = BuildMethods.field(
        verbose_name=_("Default build method"),
        blank=True, null=True)

    simulate_today = models.DateField(
        _("Simulated date"), blank=True, null=True)

    site_company = dd.ForeignKey(
        "contacts.Company",
        blank=True, null=True,
        verbose_name=_("Site owner"),
        related_name='site_company_sites')


    def __str__(self):
        return force_text(_("Site Parameters"))

    def update(self, **kw):
        """
        Set some field of the SiteConfig object and store it to the
        database.
        """
        # print("20180502 update({})".format(kw))
        for k, v in kw.items():
            if not hasattr(self, k):
                raise Exception("SiteConfig has no attribute %r" % k)
            setattr(self, k, v)
        self.full_clean()
        self.save()

    def save(self, *args, **kw):
        # print("20180502 save() {}".format(dd.obj2str(self, True)))
        super(SiteConfig, self).save(*args, **kw)
        settings.SITE.clear_site_config()


def my_handler(sender, **kw):
    # print("20180502 {} my_handler calls clear_site_config()".format(
    #     settings.SITE))
    settings.SITE.clear_site_config()
    #~ kw.update(sender=sender)
    # dd.database_connected.send(sender)
    #~ dd.database_connected.send(sender,**kw)

from django.test.signals import setting_changed
from lino.utils.djangotest import testcase_setup
setting_changed.connect(my_handler)
testcase_setup.connect(my_handler)
dd.connection_created.connect(my_handler)
models.signals.post_migrate.connect(my_handler)


class SiteConfigs(dd.Table):

    model = 'system.SiteConfig'
    required_roles = dd.login_required(SiteStaff)
    # default_action = actions.ShowDetail()
    #~ has_navigator = False
    hide_top_toolbar = True
    #~ can_delete = perms.never
    detail_layout = """
    default_build_method
    # lino.ModelsBySite
    """

    @classmethod
    def get_default_action(cls):
        return actions.ShowDetail(cls.detail_layout)



    do_build = BuildSiteCache()


# if settings.SITE.user_model == 'users.User':
#     dd.inject_field(settings.SITE.user_model,
#                     'user_type', UserTypes.field())
#     dd.inject_field(settings.SITE.user_model, 'language', dd.LanguageField())



class BleachChecker(Checker):

    verbose_name = _("Find unbleached html content")
    model = dd.Model

    def get_checkable_models(self):

        for m in super(BleachChecker, self).get_checkable_models():
            if len(m._bleached_fields):
                yield m

    def get_checkdata_problems(self, obj, fix=False):
        t = tuple(obj.fields_to_bleach())
        if len(t):
            fldnames = ', '.join([f.name for f, old, new in t])
            yield (True, _("Fields {} have unbleached content.").format(fldnames))
            if fix:
                obj.before_ui_save(None)
                obj.full_clean()
                obj.save()


BleachChecker.activate()



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

import json

from lino.api import rt
import logging
from django.utils.timezone import now
logger = logging.getLogger(__name__)

NOTIFICATION = "NOTIFICATION"
CHAT = "CHAT"

NOTIFICATION_TYPES = [
    NOTIFICATION, CHAT
]


def send_notification(user, id, subject, body, created):
    """
    :param user:
    :param id:
    :param subject:
    :param body:
    :param created:
    :return:
    """
    # importing channels at module level would cause certain things to fail
    # when channels isn't installed, e.g. `manage.py prep` in `lino_book.projects.workflows`.
    from channels.layers import get_channel_layer
    from asgiref.sync import async_to_sync

    msg = dict(
        type=NOTIFICATION,
        subject=subject,
        id=id,
        body=body,
        created=created,
    )

    try:
        channel_layer = get_channel_layer()
        async_to_sync(channel_layer.group_send)(str(user.pk),
                                                {"type": "send_notification",  # method name in consumer
                                                 "text": json.dumps(msg)})  # data

    except Exception as E:
        logger.exception(E)


def send_global_chat(message):
    """
    Sends a WS message to each user using ChatProps"""
    from channels.layers import get_channel_layer
    from asgiref.sync import async_to_sync

    for chat in message.chatProps.all():
        msg = dict(
            type=CHAT,
            chat=chat.serialize())

        try:
            assert bool(chat.user)
            channel_layer = get_channel_layer()
            async_to_sync(channel_layer.group_send)(str(chat.user.pk),
                                                    {"type": "send_notification",
                                                     # just pointer to method name in consumer
                                                     "text": json.dumps(msg)})  # data
            chat.sent = now()
            chat.save()
        except Exception as E:
            logger.exception(E)



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from __future__ import unicode_literals
# from builtins import str

from atelier.utils import isidentifier

from lino.api import dd, _, pgettext





# -*- coding: UTF-8 -*-
# Copyright 2011-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from lino.api import dd
class ChatGroups(dd.Table):
    model = 'chat.ChatGroup'
    column_names = "created user title description *"
    required_roles = set([])
    # cell_edit = False

    detail_layout = dd.DetailLayout("""
     title 
     created 
     user 
     description
     ChatMembers
    """, window_size=(50, 15))

    insert_layout = dd.InsertLayout("""
    title
    description
    """)

class ChatMembers(dd.Table):
    model = "chat.ChatGroupMember"
    master_key = "group"

# from lino.core.roles import UserRole

class ChatMessages(dd.Table):
    model = 'chat.ChatMessage'
    column_names = "created user body *"
    # required_roles = [UserRole]
    # cell_edit = False

    detail_layout = dd.DetailLayout("""
     user group 
     body
    """, window_size=(50, 15))

    #parameters = ObservedDateRange(
        # user=dd.ForeignKey(
        #     settings.SITE.user_model,
        #     blank=True, null=True),
        # show_seen=dd.YesNo.field(_("Seen"), blank=True),
    #)

    #params_layout = "user start_date end_date"

    # @classmethod
    # def get_simple_parameters(cls):
    #     for p in super(Messages, cls).get_simple_parameters():
    #         yield p
    #     yield 'user'

    @classmethod
    def get_request_queryset(self, ar, **filter):
        qs = super(ChatMessages, self).get_request_queryset(ar, **filter)
        # pv = ar.param_values
        #
        # if pv.show_seen == dd.YesNo.yes:
        #     qs = qs.filter(seen__isnull=False)
        # elif pv.show_seen == dd.YesNo.no:
        #     qs = qs.filter(seen__isnull=True)
        return qs



# -*- coding: UTF-8 -*-
# Copyright 2009-2019 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)
"""Callback handling
"""

from __future__ import unicode_literals, print_function
from builtins import object

import logging; logger = logging.getLogger(__name__)

import sys
import os
from os.path import join, dirname, exists

# from multiprocessing import Process, Manager

from django.conf import settings
from django.utils.translation import ugettext as _
from django.utils.encoding import force_text

from etgen.html import tostring
from hashlib import md5

# from lino.core.requests import ActorRequest

CLONEABLE_ATTRS = frozenset("""ah request user subst_user
bound_action create_kw known_values param_values
action_param_values""".split())


class CallbackChoice(object):
    # ~ def __init__(self,name,label,func):

    def __init__(self, name, func, label):
        self.name = name
        self.func = func
        self.label = str(label)


class Callback(object):
    """A callback is a question that rose during an AJAX action.
    The original action is pending until we get a request
    that answers the question.

    TODO: move all callback-related code out of
    :mod:`lino.core.kernel` into to a separate module and install it
    as a "kernel plugin" in a similar way as :mod:`lino.core.web` and
    :mod:`lino.utils.config`.

    """
    title = _('Confirmation')

    def __init__(self, ar, message, uid=None):
        self.message = message
        self.choices = []
        self.choices_dict = {}
        self.ar = ar

        str_msg = tostring(message).encode()
        self.uid = uid if uid is not None else str(md5(str_msg).digest())

    def __repr__(self):
        return "Callback(%r)" % self.message

    def set_title(self, title):
        self.title = title

    def add_choice(self, name, func, label):
        """
        Add a possible answer to this callback.
        - name: "yes", "no", "ok" or "cancel"
        - func: a callable to be executed when user selects this choice
        - the label of the button
        """
        assert not name in self.choices_dict
        allowed_names = ("yes", "no", "ok", "cancel")
        if not name in allowed_names:
            raise Exception("Sorry, name must be one of %s" % allowed_names)
        cbc = CallbackChoice(name, func, label)
        self.choices.append(cbc)
        self.choices_dict[name] = cbc
        return cbc

    def run(self, choice):
        if choice not in self.choices_dict:
            raise Exception("Can not run choice|%s| not in avaliable choices %s"%(choice,self.choices_dict))
        self.choices_dict[choice].func(self.ar)

class unused_CallbackManager(object):

    def __init__(self):

        self.pending_threads = {}

    def run_callback(self, request, thread_id, button_id):
        """Continue the action which was started in a previous request and
        which asked for user interaction via a :class:`Callback`.

        This is called from :class:`lino.modlib.extjs.views.Callbacks` and
        similar places in other front ends.

        """
        # logger.info("20131212 get_callback %s %s", thread_id, button_id)

        # 20140304 Also set a renderer so that callbacks can use it
        # (feature needed by beid.FindByBeIdAction).

        thread_id = thread_id
        cb = self.pending_threads.pop(thread_id, None)
        if cb is None:
            logger.debug("No callback %r in %r" % (
                thread_id, list(self.pending_threads.keys())))
            raise Warning("Unknown callback %r" % thread_id)
            # ar = ActorRequest(request, renderer=self.default_renderer)
            # ar.error("Unknown callback %r" % thread_id)
            # return ar.renderer.render_action_response(ar)

        # e.g. SubmitInsertClient must set `data_record` in the
        # callback request ("ar2"), not the original request ("ar"),
        # i.e. the methods to create an instance and to fill
        # `data_record` must run on the callback request.  So the
        # callback request must be a clone of the original request.
        # New since 20140421
        ar = cb.ar.actor.request_from(cb.ar)
        for k in CLONEABLE_ATTRS:
            setattr(ar, k, getattr(cb.ar, k))

        for c in cb.choices:
            if c.name == button_id:
                a = ar.bound_action.action
                if settings.SITE.log_each_action_request and not a.readonly:
                    logger.info("run_callback {0} {1} {2}".format(
                        thread_id, cb.message, c.name))
                    # ar.info(cb.message)
                func = c.func
                try:
                    func(ar)
                except Warning as e:
                    ar.error(e, alert=True)
                return ar.renderer.render_action_response(ar)

        ar.error("Invalid button %r for callback %r" % (button_id, thread_id))
        return ar.renderer.render_action_response(ar)

def add_callback(ar, *msgs, uid=None):
    """
    Returns an *action callback* which will initiate a dialog thread by
    asking a question to the user and suspending execution until
    the user's answer arrives in a next HTTP request.

    Calling this from an Action's :meth:`run_from_ui
    <lino.core.actions.Action.run_from_ui>` method will interrupt
    the execution, send the specified message back to the client.

    The client will display the prompt and do another request on the same action
    but with the given answer.

    """
    if len(msgs) > 1:
        msg = '\n'.join([force_text(s) for s in msgs])
    else:
        msg = msgs[0]
    # logger.info("20160526 add_callback(%s)", msg)
    return Callback(ar, msg, uid=uid)

def set_callback(ar, cb):
    """
    Runs when confimation is requred, checks if answer is present in request data, if so, runs corisponding function
    Otherwise returns with data that will pop up a dialog.
    """
    cb_id = cb.uid
    answer = ar.xcallback_answers.get("xcallback__" + cb_id, None)
    if answer is None:
        buttons = []
        xcallback = dict(
            id=cb_id,
            title=cb.title,
            buttons=buttons)

        rq_data = { k:v[0] if len(v) == 1 else v
            for k,v in (ar.rqdata.lists() if getattr(ar, "rqdata", None) is not None else {})}
        rq_data.pop("_dc", None)

        for c in cb.choices:
            buttons.append([c.name, c.label])
            xcallback[c.name + "_resendEvalJs"] = ar.renderer.ar2js(
                ar, ar.selected_rows, rqdata= rq_data, xcallback={
                    "xcallback_id" : cb_id,
                    "choice" : c.name })
        # print(buttons)
        return ar.success(cb.message, xcallback=xcallback)

    else:
        cb.run(answer)

def popCallBack(resp):
    """
    :param resp: a dict which has some callback choices in it
    :return: dict of all callback choices
    """
    return {cb:c for cb,c in resp.items() if cb.startswith("xcallback__")}

def applyCallbackChoice(resp = {}, data={}, choice="yes"):
    """
    Looks up all other callback IDs from a previous responce as well as adds your new choice to the data for resending.
    Used in testing
    """
    data.update(popCallBack(resp))
    callback = {"xcallback__" + resp["xcallback"]['id'] : choice}
    data.update(callback)
    return data
# mpman = Manager()
# mgr = CallbackManager()
#
# set_callback = mgr.set_callback
# add_callback = mgr.add_callback
# run_callback = mgr.run_callback



# -*- coding: UTF-8 -*-
# Copyright 2015-2019 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from django.db import models
from django.db.models import Q

from lino.api import dd, _

from .roles import PrivateCommentsReader

class Commentable(dd.Model):
    class Meta(object):
        abstract = True

    def on_commented(self, comment, ar, cw):
        pass

    def get_rfc_description(self, ar):
        return ''

    def get_comment_group(self):
        return None

    @classmethod
    def get_comments_filter(cls, user):
        if user.user_type.has_required_roles([PrivateCommentsReader]):
            return None
        if user.is_anonymous:
            return Q(private=False)
        return Q(private=False) | Q(user=user)

    def is_comment_private(self, comment, ar):
        """Whether the given comment should be private."""
        return dd.plugins.comments.private_default



# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

"""Adds functionality for instant messageing through websockets.

See :doc:`/specs/chat`.

"""

from lino.api import ad, _
from lino.core.utils import is_devserver
try:
    import redis
except ImportError:
    redis = None
try:
    import channels
except ImportError:
    channels = None


class Plugin(ad.Plugin):
    "See :class:`lino.core.plugin.Plugin`."

    verbose_name = _("Chat")

    needs_plugins = ['lino.modlib.notify','lino_xl.lib.groups']
    #extends_models = ['Group']

    media_name = 'js'

    def on_init(self):
        if not self.site.use_websockets:
            raise Warning("Chat requires use_websockets = True")

    # def get_requirements(self, site):
    #     # handled by modlib.notify
    #     pass

    # def get_used_libs(self, html=None):
    #     handled by modlib.notify
        # pass


    def setup_main_menu(self, site, user_type, m):
    #     todo,
        # pass

        p = site.plugins.office
        m = m.add_menu(p.app_label, p.verbose_name)
        m.add_action('chat.ChatMessages')

    def setup_explorer_menu(self, site, user_type, m):
        # todo,
        #pass
        p = site.plugins.system
        m = m.add_menu(p.app_label, p.verbose_name)
        m.add_action('chat.ChatMessages')
        m.add_action('chat.ChatGroups')


    def get_head_lines(self, site, request):
        """Adds a JS constant to let react know we want to include WS chat info"""
        if not self.site.use_websockets:
            yield ""
        # from lino.utils.jsgen import py2js
        else :
            yield """
        <script type="text/javascript">
            window.Lino = window.Lino || {}
            window.Lino.useChats = true;
        </script>
            """

    # def get_dashboard_items(self, user):
    #     if user.authenticated:
            # yield ActorItem(
            #     self.models.notify.MyMessages, header_level=None)
            # yield self.site.models.notify.MyMessages



# -*- coding: UTF-8 -*-
# Copyright 2016-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

import os
from copy import copy

try:
    from weasyprint import HTML
except ImportError:
    HTML = None

from django.conf import settings
from django.utils import translation

from lino.api import dd

from lino.modlib.jinja.choicelists import JinjaBuildMethod
from lino.modlib.printing.choicelists import BuildMethods



class WeasyBuildMethod(JinjaBuildMethod):

    template_ext = '.weasy.html'
    templates_name = 'weasy'
    default_template = 'default.weasy.html'


class WeasyHtmlBuildMethod(WeasyBuildMethod):
    target_ext = '.html'
    name = 'weasy2html'

    def html2file(self, html, filename):
        open(filename, 'w').write(html)


class WeasyPdfBuildMethod(WeasyBuildMethod):
    target_ext = '.pdf'
    name = 'weasy2pdf'

    def html2file(self, html, filename):
        pdf = HTML(string=html)
        pdf.write_pdf(filename)


add = BuildMethods.add_item_instance
add(WeasyHtmlBuildMethod())
add(WeasyPdfBuildMethod())



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)


import os
from copy import copy

from django.db import models
from django.conf import settings
from django.utils import translation

from lino.api import dd, _
from lino.mixins.registrable import RegistrableState

from lino.modlib.jinja.choicelists import JinjaBuildMethod
from lino.modlib.printing.choicelists import BuildMethods

class PublisherBuildMethod(JinjaBuildMethod):

    template_ext = '.pub.html'
    templates_name = 'pub'
    default_template = 'default.pub.html'
    target_ext = '.html'
    name = 'pub'

    def build(self, ar, action, elem):
        filename = action.before_build(self, elem)
        if filename is None:
            return
        tpl = self.get_template(action, elem)
        lang = str(elem.get_print_language()
                   or settings.SITE.DEFAULT_LANGUAGE.django_code)
        ar = copy(ar)
        ar.renderer = settings.SITE.plugins.jinja.renderer
        # ar.tableattrs = dict()
        # ar.cellattrs = dict(bgcolor="blue")

        with translation.override(lang):
            cmd_options = elem.get_build_options(self)
            dd.logger.info(
                "%s render %s -> %s (%r, %s)",
                self.name, tpl.lino_template_names, filename, lang, cmd_options)
            context = elem.get_printable_context(ar)
            html = tpl.render(context)
            self.html2file(html, filename)
            return os.path.getmtime(filename)

    def html2file(self, html, filename):
        raise NotImplementedError()


BuildMethods.add_item_instance(PublisherBuildMethod())



class PublishableState(RegistrableState):
    is_published = False


class PublishableStates(dd.Workflow):
    item_class = PublishableState
    verbose_name = _("Publishable state")
    verbose_name_plural = _("Publishable states")
    column_names = "value name text is_published"

    @classmethod
    def get_published_states(cls):
        return [o for o in cls.objects() if o.is_published]

    @dd.virtualfield(models.BooleanField(_("published")))
    def is_published(cls, choice, ar):
        return choice.is_published

add = PublishableStates.add_item
add('10', _("Draft"), 'draft')
add('20', _("Published"), 'published', is_published=True)
add('30', _("Removed"), 'removed')



# -*- coding: UTF-8 -*-
# Copyright 2016-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)


import os
from copy import copy

from django.conf import settings
from django.utils import translation

from lino.api import dd
from lino.modlib.printing.choicelists import DjangoBuildMethod


class JinjaBuildMethod(DjangoBuildMethod):

    def build(self, ar, action, elem):
        filename = action.before_build(self, elem)
        if filename is None:
            return
        tpl = self.get_template(action, elem)
        lang = str(elem.get_print_language()
                   or settings.SITE.DEFAULT_LANGUAGE.django_code)
        ar = copy(ar)
        ar.renderer = settings.SITE.plugins.jinja.renderer
        # ar.tableattrs = dict()
        # ar.cellattrs = dict(bgcolor="blue")

        with translation.override(lang):
            cmd_options = elem.get_build_options(self)
            dd.logger.info(
                "%s render %s -> %s (%r, %s)",
                self.name, tpl.lino_template_names, filename, lang, cmd_options)
            context = elem.get_printable_context(ar)
            html = tpl.render(context)
            self.html2file(html, filename)
            return os.path.getmtime(filename)

    def html2file(self, html, filename):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Copyright 2012-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)
"""Some diagnostic utilities."""

# from textwrap import fill
from atelier import rstgen
from atelier.utils import unindent

from django.conf import settings
from django.db import models
from django.utils.encoding import force_text

from lino.modlib.system.choicelists import PeriodEvents
from lino.core.layouts import BaseLayout
from lino.core.layouts import ParamsLayout
from lino.core.fields import DummyField
from lino.core.elems import Container, Wrapper, FieldElement
from lino.modlib.users.choicelists import UserTypes
from lino.core import actors
from lino.core import actions
from lino.core.utils import get_models, sorted_models_list
from lino.core.utils import full_model_name as fmn
from lino.api import dd


class Analyzer(object):
    "The class of the :data:`lino.utils.diag.analyzer` object."
    def __init__(self):
        self._initialized = False

    def analyze(self):
        if self._initialized:
            return
        self._initialized = True
        window_actions = dict()
        self.custom_actions = []
        for a in actors.actors_list:
            if a.abstract:
                continue
            for ba in a.get_actions():
                if ba.action.is_window_action():
                    wl = ba.get_window_layout() or ba.action.params_layout
                    if wl is not None:
                        if isinstance(wl, str):
                            raise Exception("20150323 : {0}".format(ba))
                            # Was used to find Exception: 20150323 :
                            # <BoundAction(checkdata.Checkers,
                            # <ShowDetail detail (u'Detail')>)>

                        if wl not in window_actions:
                            # lh = wl.get_layout_handle(ui)
                            # for e in lh.main.walk():
                            #     e.loosen_requirements(a)
                            window_actions[wl] = ba
                else:  # if ba.action.custom_handler:
                    self.custom_actions.append(ba)
        l = list(window_actions.values())

        def f(a):
            return str(a.full_name())
        self.window_actions = list(sorted(l, key=f))
        self.custom_actions = list(sorted(self.custom_actions, key=f))

    def show_window_fields(self):
        """List all window actions and the form fields they contain.
        """
        self.analyze()
        items = []
        for ba in analyzer.window_actions:
            items.append(
                "{0} : {1}".format(
                    ba.full_name(), layout_fields(ba)))

        return rstgen.ul(items)

    def show_window_permissions(self):
        self.analyze()
        items = []
        for ba in analyzer.window_actions:
            items.append(
                "{0} : visible for {1}".format(
                    ba.full_name(), visible_for(ba)))

        return rstgen.ul(items)

    def show_memo_commands(self, doctestfmt=False):
        rst = ""
        mp = settings.SITE.plugins.memo.parser
        items = []
        for cmd, func in sorted(mp.commands.items()):
            doc = unindent(func.__doc__ or '')
            if doc:
                # doc = doc.splitlines()[0]
                items.append(
                    "[{0} ...] : {1}".format(cmd, doc))

        # rst += "\n**Commands**"
        # rst += rstgen.boldheader("Commands")
        rst += rstgen.ul(items)

        if False:
            items = []
            for model, func in sorted(mp.renderers.items()):
                doc = unindent(func.__doc__ or '')
                if doc:
                    items.append(
                        "[{0} ...] : {1}".format(model, doc))
            if len(items):
                rst += "\n**Renderers**"
                rst += rstgen.ul(items)


        return rst

    def show_dialog_actions(self, doctestfmt=False):
        self.analyze()
        items = []
        for ba in analyzer.custom_actions + analyzer.window_actions:
            # if ba.action.parameters and not ba.action.no_params_window:
            if ba.action.parameters:
                items.append(
                    "{0} : {1}".format(
                        ba.full_name(),
                        py2rst(ba.action, doctestfmt)))

        print(rstgen.ul(items))

    def show_action_permissions(self, *classes):
        self.analyze()
        items = []
        for ba in analyzer.custom_actions + analyzer.window_actions:
            if isinstance(ba.action, classes):
                items.append(
                    "{0} : visible for {1}".format(
                        ba.full_name(), visible_for(ba)))

        return rstgen.ul(items)

    def show_database_structure(self):
        """Show a bullet list of all models and their fields."""
        self.analyze()
        items = []
        for model in get_models():
            names = []
            # for f, m in model._meta.get_fields_with_model():
            for f in model._meta.concrete_fields:
                names.append(f.name)
            items.append(
                "{0} : {1}".format(fmn(model), ', '.join(names)))

        items = sorted(items)
        return rstgen.ul(items)

    def show_fields(self, model, field_names=None, languages=None):
        model = dd.resolve_model(model)
        if field_names is not None:
            field_names = dd.fields_list(model, field_names)
        items = []
        for f in model._meta.fields:
            if field_names is None or f.name in field_names:
                name = f.name
                ref = model.__module__ + '.' + model.__name__ + '.' + name
                verbose_name = force_text(f.verbose_name).strip()
                help_text = force_text(f.help_text).replace('\n', ' ')
                txt = "**{verbose_name}** (:attr:`{name} <{ref}>`) : " \
                      "{help_text}".format(**locals())
                items.append(txt)
        return rstgen.ul(items)

    def show_db_overview(self):
        """Return a reStructredText-formatted "database overview" report.
        Used by test cases in tested documents.

        """

        models_list = sorted_models_list()
        apps = [p.app_label for p in settings.SITE.installed_plugins]
        s = "%d apps: %s." % (len(apps), ", ".join(apps))
        s += "\n%d models:\n" % len(models_list)
        i = 0
        headers = [
            #~ "No.",
            "Name",
            "Default table",
            #~ "M",
            "#fields",
            "#rows",
            #~ ,"first","last"
        ]
        rows = []
        for model in models_list:
            if True:  # model._meta.managed:
                i += 1
                cells = []
                #~ cells.append(str(i))
                cells.append(fmn(model))
                cells.append(model.get_default_table())
                #~ cells.append(str(model))
                #~ if model._meta.managed:
                #~ cells.append('X')
                #~ else:
                #~ cells.append('')
                cells.append(str(len(model._meta.concrete_fields)))
                qs = model.objects.all()
                n = qs.count()
                cells.append(str(n))
                #~ if n:
                #~ cells.append(obj2str(qs[0]))
                #~ cells.append(obj2str(qs[n-1]))
                #~ else:
                #~ cells.append('')
                #~ cells.append('')

                rows.append(cells)
        s += rstgen.table(headers, rows)
        return s

    def show_foreign_keys(self):
        """Return a list that shows how database objects are being referred to
        by some other database object. This information is important
        (1) before deleting objects and (2) when merging them.

        For every model we see a list of "delete handlers" and a list
        of fields from other models that point to this model using
        that delete handler.

        Delete handlers are:

        - PROTECT : refuse to delete when other objects refer to this object
        - CASCADE : delete objects refering to this object
        - set_on_delete : make other objects point to something else (or set
          their pointer to None)

        """
        self.analyze()
        tdp = dict()  # target model -> delete handler -> pointer list
        for target in get_models():
            dp = tdp.setdefault(target, dict())
            for m, fk in target._lino_ddh.fklist:
                k = fk.remote_field.on_delete
                p = dp.setdefault(k, [])
                p.append((m, fk))

        def fk2str(mfk):
            return "{0}.{1}".format(fmn(mfk[0]), mfk[1].name)

        items1 = []
        for target, dp in list(tdp.items()):
            items2 = []
            for dh, pl in list(dp.items()):
                items2.append(
                    "{0} : {1}".format(
                        dh.__name__, ', '.join([fk2str(mfk) for mfk in pl])))
            if len(items2):
                items2 = sorted(items2)
                items1.append("{0} :\n{1}".format(
                    fmn(target), rstgen.ul(items2)))

        items1 = sorted(items1)
        return rstgen.ul(items1)

    def get_complexity_factors(self, today=None):
        self.analyze()
        yield "{0} plugins".format(len(dd.plugins))
        yield "{0} models".format(len(get_models()))
        User = settings.SITE.user_model
        if today and User:
            qs = User.objects.filter(username__isnull=False)
            qs = PeriodEvents.active.add_filter(qs, today)
            yield "{0} users".format(qs.count())
        yield "{0} user roles".format(len(settings.SITE.user_roles))
        yield "{0} user types".format(len(UserTypes.objects()))
        yield "{0} views".format(len(
            [a for a in actors.actors_list if not a.abstract]))
        dialog_actions = [ba for ba in analyzer.custom_actions +
                          analyzer.window_actions
                          if ba.action.parameters and not ba.action.no_params_window]
        yield "{0} dialog actions".format(len(dialog_actions))

    def show_complexity_factors(self):
        return rstgen.ul(list(self.get_complexity_factors()))


def visible_for(ba):
    """Shows a list of user profiles for which this action is visible."""
    if ba is None:
        return "N/A"
    if isinstance(ba, type)  and issubclass(ba, actors.Actor):
        ba = ba.default_action
    visible = []
    hidden = []
    for p in UserTypes.objects():
        name = p.name or p.value
        if ba.get_view_permission(p):
            visible.append(name)
        else:
            hidden.append(name)
    if len(hidden) == 0:
        return "all"
    if len(visible) == 0:
        return "nobody"
    # if len(hidden) < len(visible):
    # if len(hidden) <= 3:
    #     return "all except %s" % ', '.join(hidden)
    return ' '.join(visible)


def layout_fields(ba):
    wl = ba.get_window_layout() or ba.action.params_layout
    if wl is None:
        return ''
    lh = wl.get_layout_handle(settings.SITE.kernel.default_ui)
    elems = [str(f.name) for f in lh._store_fields if not isinstance(f, DummyField)]
    return ', '.join(elems)
    # return fill(' '.join(elems), 50)

def elem_label(e):
    if isinstance(e, FieldElement):
        return "**%s** (%s)" % (str(e.field.verbose_name), e.field.name)
    elif e.get_label() is None:
        return "(%s)" % e.name
    else:
        return "**%s** (%s)" % (str(e.get_label()), e.name)


def py2rst(self, doctestfmt=False, fmt=None):
    """
    Return a textual representation of the given Python object as a
    reStructuredText bullet list.

    The Python object can be a layout, a layout element, an action or a
    database object.

    If it is an action, it must have parameters, and py2rst will render the
    params_layout.

    If it is a database object, you will get  a textual representation of a
    detail window on that object.

    If the optional argument `doctestfmt` is specified as `True`, then
    output contains less blank lines, which might be invalid
    reStructuredText but is more doctest-friendly.

    TODO: move this functionality to lino.api.doctests and rename it to
    something that reflects better what it does.

    """
    from lino.core.store import get_atomizer

    if isinstance(self, models.Model) and fmt is None:
        ar = self.get_default_table().request()
        def fmt(e):
            s = elem_label(e)
            if isinstance(e, FieldElement):
                sf = get_atomizer(self.__class__, e.field, e.field.name)
                getter = sf.full_value_from_object
                value = getter(self, ar)
                # value = e.value_from_object(self, None)
                if value is not None:
                    s += ": " + e.format_value(ar, value)
            return s

        dt = self.get_default_table()
        lh = dt.detail_layout.get_layout_handle()
        return py2rst(lh.main, doctestfmt, fmt)

    if isinstance(self, actions.Action):
        s = str(self)
        if self.params_layout:
            lh = self.params_layout.get_layout_handle(
                settings.SITE.kernel.default_ui)
            s += '\n'
            s += py2rst(lh.main, doctestfmt, fmt)
        return s


    if isinstance(self, BaseLayout):
        lh = self.get_layout_handle(settings.SITE.kernel.default_ui)
        return py2rst(lh.main, doctestfmt, fmt)

    if isinstance(self, Wrapper):
        self = self.wrapped

    if fmt is None:
        def fmt(e):
            s = elem_label(e)
            if visible_for(e) != visible_for(e.parent):
                s += " [visible for %s]" % visible_for(e)
            return s

    s = fmt(self)
    if isinstance(self, Container):
        use_ul = False
        for e in self.elements:
            if isinstance(e, Container):
                use_ul = True
        children = [py2rst(e, doctestfmt, fmt) for e in self.elements]
        if len(children):
            if use_ul:
                s += ':\n'
                if not doctestfmt:
                    s += '\n'
                s += rstgen.ul(children)
            else:
                s += ": " + ', '.join(children)

    return s


analyzer = Analyzer()
"""This is a docstring
"""



# -*- coding: UTF-8 -*-
# Copyright 2010-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

import logging ; logger = logging.getLogger(__name__)

from django.conf import settings
from lino.modlib.users.choicelists import UserTypes


def root_user(lang, **kw):
    # ~ kw.update(user_type='900') # UserTypes.admin)
    #~ print 20130219, UserTypes.items()
    kw.update(user_type=UserTypes.admin)
    kw.update(email=settings.SITE.demo_email)  # 'root@example.com'
    lang = lang.django_code
    kw.update(language=lang)
    lang = lang[:2]
    if lang == 'de':
        kw.update(first_name="Rolf", last_name="Rompen")
    elif lang == 'fr':
        kw.update(first_name="Romain", last_name="Raffault")
    elif lang == 'et':
        kw.update(first_name="Rando", last_name="Roosi")
    elif lang == 'en':
        kw.update(first_name="Robin", last_name="Rood")
    elif lang == 'pt':
        kw.update(first_name="Ronaldo", last_name="Rosa")
    elif lang == 'es':
        kw.update(first_name="Rodrigo", last_name="Rosalez")
    elif lang == 'nl':
        kw.update(first_name="Rik", last_name="Rozenbos")
    else:
        logger.warning("No demo user for language %r.", lang)
        return None
    kw.update(username=kw.get('first_name').lower())
    return kw


def objects():
    # logger.info("20150323 %s", settings.SITE.languages)
    User = settings.SITE.user_model
    if User is not None:
        for lang in settings.SITE.languages:
            if settings.SITE.hidden_languages is None \
               or not lang.django_code in settings.SITE.hidden_languages:
                kw = root_user(lang)
                if kw:
                    yield User(**kw)



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)


import logging ; logger = logging.getLogger(__name__)

from django.views.generic import View





class Element(View):
    
    publisher_model = None

    def get(self, request, pk=None):
        obj = self.publisher_model.objects.get(id=pk)
        return obj.get_publisher_response(request)



# -*- coding: UTF-8 -*-
# Copyright 2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from lino.api.ad import Plugin

from lino.core.dashboard import DashboardItem

class PublisherDashboardItem(DashboardItem):

    def __init__(self, obj, **kwargs):
        self.obj = obj
        super(PublisherDashboardItem, self).__init__(None, **kwargs)

    def render(self, ar):
        return obj.get_publisher_response(ar)

class Plugin(Plugin):

    needs_plugins = ["lino.modlib.jinja"]

    def get_patterns(self):
        from django.conf.urls import url
        from lino.core.utils import models_by_base
        from . import views
        from .mixins import Publishable

        for m in models_by_base(Publishable):
            if m.publisher_location is not None:
                yield url('^{}/(?P<pk>.+)$'.format(m.publisher_location),
                    views.Element.as_view(publisher_model=m))

    def get_dashboard_items(self, user):
        from lino.core.utils import models_by_base
        from .mixins import Publishable
        for m in models_by_base(Publishable):
            if m.publisher_location is not None:
                for obj in m.get_dashboard_objects(user):
                    yield PublisherDashboardItem(obj)



# -*- coding: UTF-8 -*-
# Copyright 2013-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)


from django.db import models
from django.db.models import Q
from django.contrib.contenttypes.models import ContentType
from django.conf import settings

from lino.api import dd, rt, _

from lino.mixins import CreatedModified, BabelNamed
from lino.modlib.users.mixins import UserAuthored
from lino.modlib.notify.mixins import ChangeNotifier
from lino.modlib.gfks.mixins import Controllable
from lino.modlib.memo.mixins import Previewable
from lino.modlib.publisher.mixins import Publishable
from .choicelists import CommentEvents
from .mixins import Commentable
# from .choicelists import PublishAllComments, PublishComment



if dd.is_installed("inbox"):
    from lino_xl.lib.inbox.models import comment_email

class CommentType(BabelNamed):
    class Meta(object):
        abstract = dd.is_abstract_model(__name__, 'CommentType')
        verbose_name = _("Comment Type")
        verbose_name_plural = _("Comment Types")



class Comment(CreatedModified, UserAuthored, Controllable,
              ChangeNotifier, Previewable, Publishable):
    class Meta(object):
        app_label = 'comments'
        abstract = dd.is_abstract_model(__name__, 'Comment')
        verbose_name = _("Comment")
        verbose_name_plural = _("Comments")

    publisher_location = "c"

    reply_to = dd.ForeignKey(
        'self', blank=True, null=True, verbose_name=_("Reply to"))
    # more_text = dd.RichTextField(_("More text"), blank=True)

    private = models.BooleanField(
        _("Private"), default=dd.plugins.comments.private_default)

    comment_type = dd.ForeignKey('comments.CommentType', blank=True, null=True)

    def __str__(self):
        return u'%s #%s' % (self._meta.verbose_name, self.pk)
        # return _('{user} {time}').format(
        #     user=self.user, obj=self.owner,
        #     time=naturaltime(self.modified))

    @classmethod
    def get_user_queryset(cls, user):
        qs = super(Comment, cls).get_user_queryset(user)

        if not user.user_type.has_required_roles([CommentsReader]):
            return qs.none()

        filters = []
        for m in rt.models_by_base(Commentable):
            flt = m.get_comments_filter(user)
            if flt is not None:
                ct = rt.models.contenttypes.ContentType.objects.get_for_model(m)
                filters.append(flt | ~Q(owner_type=ct))
        if len(filters):
            qs = qs.filter(*filters)
        return qs.distinct()  # add distinct because filter might be on a join

    # def after_ui_create(self, ar):
    #     super(Comment, self).after_ui_create(ar)
    #     if self.owner_id:
    #         self.private = self.owner.is_comment_private(self, ar)

    def on_create(self, ar):
        super(Comment, self).on_create(ar)
        if self.owner_id:
            self.private = self.owner.is_comment_private(self, ar)

    def after_ui_save(self, ar, cw):
        super(Comment, self).after_ui_save(ar, cw)
        if self.owner_id:
            self.owner.on_commented(self, ar, cw)
        if dd.is_installed("memo"):
            ref_objects = settings.SITE.plugins.memo.parser.get_referred_objects(self.body)
            for ref_object in ref_objects:
                created_mention = Mention(comment=self,
                        owner_id=ref_object.pk,
                        owner_type=ContentType.objects.get_for_model(ref_object.__class__))
                created_mention.touch()
                created_mention.save()

    # def full_clean(self):
    #     super(Comment, self).full_clean()
    #     self.owner.setup_comment(self)

    def get_change_owner(self):
        return self.owner or self

    # def get_change_message_type(self, ar):
    #     if self.published is None:
    #         return None
    #     return super(Comment, self).get_change_message_type(ar)

    def get_change_observers(self, ar=None):
        if isinstance(self.owner, ChangeNotifier):
            obs = self.owner
        else:
            obs = super(Comment, self)
        for u in obs.get_change_observers(ar):
            yield u

    def get_change_subject(self, ar, cw):
        if cw is None:
            s = _("{user} commented on {obj}")
        else:
            s = _("{user} modified comment on {obj}")
        return s.format(user=ar.get_user(), obj=self.owner)

    def get_change_body(self, ar, cw):
        if cw is None:
            s = _("{user} commented on {obj}")
        else:
            s = _("{user} modified comment on {obj}")
        user = ar.get_user()
        s = s.format(
            user=user, obj=ar.obj2memo(self.owner))
        if dd.is_installed("inbox"):
            #mailto:ADDR@HOST.com?subject=SUBJECT&body=Filling%20in%20the%20Body!%0D%0Afoo%0D%0Abar
            s += ' <a href="{href}">{reply}</a>'.format(
                href=comment_email.gen_href(self, user),
                reply=_("Reply"))

        s += ':<br>' + self.body
        # if False:
        #     s += '\n<p>\n' + self.more_text
        return s

    @classmethod
    def setup_parameters(cls, fields):
        fields.update(
            observed_event=CommentEvents.field(blank=True))
        fields.update(
            start_date=models.DateField(
                _("Period from"), blank=True, null=True,
                help_text=_("Start date of observed period")))
        fields.update(
            end_date=models.DateField(
                _("until"),
                blank=True, null=True,
                help_text=_("End date of observed period")))
        # fields.update(
        #     show_published=dd.YesNo.field(_("Published"), blank=True))
        super(Comment, cls).setup_parameters(fields)

    @classmethod
    def get_request_queryset(cls, ar, **filter):
        qs = super(Comment, cls).get_request_queryset(ar, **filter)
        pv = ar.param_values
        if pv.observed_event:
            qs = pv.observed_event.add_filter(qs, pv)
        # if pv.show_published == dd.YesNo.yes:
        #     qs = qs.filter(published__isnull=False)
        # elif pv.show_published == dd.YesNo.no:
        #     qs = qs.filter(published__isnull=True)
        return qs

dd.update_field(Comment, 'user', editable=False)
Comment.update_controller_field(verbose_name=_('Topic'))

class Mention(CreatedModified, Controllable, UserAuthored):

    class Meta(object):
        app_label = 'comments'
        abstract = dd.is_abstract_model(__name__, 'Mention')
        verbose_name = _("Mention")
        verbose_name_plural = _("Mentions")

    comment = dd.ForeignKey('comments.Comment', blank=True, null=True)

from .ui import *



# -*- coding: UTF-8 -*-
# Copyright 2009-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)


import os
import datetime

from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import make_aware

from lino.api import rt
from lino.modlib.checkdata.choicelists import Checker

from lino.utils.choosers import chooser
from lino.core.model import Model
from lino.mixins.duplicable import Duplicable


from .choicelists import BuildMethods
from .actions import (DirectPrintAction, CachedPrintAction,
                      ClearCacheAction, EditTemplate)


def weekdays(d):
    for i in range(5):
        yield d
        d += datetime.timedelta(days=1)

def register_build_method(bm):
    # not used.
    BuildMethods.add_item_instance(bm)


def build_method_choices():
    return BuildMethods.choices


class PrintableType(Model):

    templates_group = None

    class Meta(object):
        abstract = True

    build_method = BuildMethods.field(blank=True, null=True)
    template = models.CharField(_("Template"), max_length=200, blank=True)

    @classmethod
    def get_template_groups(cls):
        """Note that `get_template_groups` is a **class method** on
        `PrintableType` but an **instance method** on `Printable`.

        """
        return [str(cls.templates_group)]  # or full_model_name(cls)

    @chooser(simple_values=True)
    def template_choices(cls, build_method):
        return cls.get_template_choices(
            build_method,
            cls.get_template_groups())

    @classmethod
    def get_template_choices(cls, build_method, template_groups):
        if not build_method:
            build_method = BuildMethods.get_system_default()
        if not build_method:
            return []
        return rt.find_template_config_files(
            build_method.template_ext, *template_groups)


class Printable(Model):

    class Meta(object):
        abstract = True

    @classmethod
    def get_printable_demo_objects(cls):

        qs = cls.objects.all()
        if qs.count() > 0:
            yield qs[0]

    @classmethod
    def get_template_group(cls):
        # used by excerpts and printable
        return cls._meta.app_label + '/' + cls.__name__

    def get_body_template(self):
        return ''

    # def get_excerpt_type(self):
    #     "Return the primary ExcerptType for the given model."
    #     ContentType = settings.SITE.models.contenttypes.ContentType
    #     ct = ContentType.objects.get_for_model(
    #         self.__class__)
    #     return self.__class__.objects.get(primary=True, content_type=ct)

    def get_excerpt_options(self, ar, **kw):
        return kw

    def get_print_language(self):
        # same as EmptyTableRow.get_print_language
        return settings.SITE.DEFAULT_LANGUAGE.django_code

    def get_template_groups(self):
        return [self.__class__.get_template_group()]

    def get_print_templates(self, bm, action):
        return [bm.get_default_template(self)]

    def get_default_build_method(self):
        return BuildMethods.get_system_default()

    def get_build_method(self):
        # TypedPrintable  overrides this
        return self.get_default_build_method()

    def get_build_options(self, bm, **opts):
        # header_center
        return opts

    def get_printable_context(self, ar=None, **kw):
        # same as lino.utils.report.EmptyTableRow.get_printable_context
        if ar is not None:
            kw = ar.get_printable_context(**kw)
        kw.update(this=self)  # for backward compatibility
        kw.update(obj=self)  # preferred in new templates
        kw.update(language=self.get_print_language() or \
                  settings.SITE.DEFAULT_LANGUAGE.django_code)
        kw.update(site=settings.SITE)
        kw.update(weekdays=weekdays)
        return kw

    def before_printable_build(self, bm):
        pass


class CachedPrintable(Duplicable, Printable):

    class Meta(object):
        abstract = True

    do_print = CachedPrintAction()
    do_clear_cache = ClearCacheAction()
    edit_template = EditTemplate()

    build_time = models.DateTimeField(
        _("build time"), null=True, editable=False)

    build_method = BuildMethods.field(blank=True, null=True)

    def full_clean(self, *args, **kwargs):
        if not self.build_method:
            self.build_method = self.get_default_build_method()
        super(CachedPrintable, self).full_clean(*args, **kwargs)

    def on_duplicate(self, ar, master):
        super(CachedPrintable, self).on_duplicate(ar, master)
        self.build_time = None
        self.build_method = None

    def get_target_name(self):
        if self.build_time:
            return self.get_build_method().get_target_name(
                self.do_print, self)

    def get_build_method(self):
        return self.build_method or self.get_default_build_method()

    def get_target_url(self):
        return self.build_method.get_target_url(
            self.do_print, self)

    def get_cache_mtime(self):
        """Return the modification time (a `datetime`) of the generated cache
        file, or `None` if no such file exists.

        """
        filename = self.get_target_name()
        if not filename:
            return None
        try:
            t = os.path.getmtime(filename)
        except OSError:
            return None
        return datetime.datetime.fromtimestamp(t)

    def clear_cache(self):
        self.build_time = None
        self.save()

    def build_target(elem, ar):
        bm = elem.get_build_method()
        t = bm.build(ar, elem.__class__.do_print, elem)
        if t is None:
            raise Exception("%s : build() returned None?!")
        # t is a file timestamp as returned by os.path.getmtime()
        # expressend as the number of seconds since the epoch.
        t = datetime.datetime.fromtimestamp(t)
        if settings.USE_TZ:
            t = make_aware(t)
        elem.build_time = t
        elem.save()


class TypedPrintable(CachedPrintable):

    type = None

    class Meta(object):
        abstract = True

    def get_printable_type(self):
        return self.type

    def get_template_groups(self):
        ptype = self.get_printable_type()
        if ptype is None:
            return super(TypedPrintable, self).get_template_groups()
        return ptype.get_template_groups()

    def get_default_build_method(self):
        ptype = self.get_printable_type()
        if ptype and ptype.build_method:
            return ptype.build_method
        return super(TypedPrintable, self).get_default_build_method()

    # def get_build_method(self):
    #     if not self.build_method:
    #         return self.get_default_build_method()
    #     return self.build_method
        # ptype = self.get_printable_type()
        # if ptype and ptype.build_method:
        #     return ptype.build_method
        # return super(TypedPrintable, self).get_build_method()

    def get_print_templates(self, bm, action):
        ptype = self.get_printable_type()
        if ptype is None:
            return super(TypedPrintable, self).get_print_templates(bm, action)

        if ptype.template:
            return [ptype.template]
        return [bm.get_default_template(self)]


class CachedPrintableChecker(Checker):
    model = CachedPrintable
    verbose_name = _("Check for missing target files")

    def get_checkdata_problems(self, obj, fix=False):
        if obj.build_time is not None:
            t = obj.get_cache_mtime()
            if t is None:
                msg = _("Seems to have been built ({bt}), "
                        "but cache file is missing .")
                params = dict(bt=obj.build_time)
                yield (False, msg.format(**params))
                # if fix:
                #     obj.build_time = None
                #     obj.full_clean()
                #     obj.save()

CachedPrintableChecker.activate()



# -*- coding: UTF-8 -*-
# Copyright 2011-2020 Rumma & Ko Ltd
# License: BSD (see file COPYING for details)

from lino.modlib.notify.api import send_notification, send_global_chat
from lino.utils.format_date import fds
from lino.modlib.office.roles import OfficeUser
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.gfks.mixins import Controllable
from lino.modlib.memo.mixins import Previewable
from lino.mixins import Created, ObservedDateRange, BabelNamed, Referrable
# from lino_noi.lib.groups.models import Group
from lino.core.site import html2text
from lino.core.requests import BaseRequest
from lino.core.gfks import gfk2lookup
from lino.api import dd, rt, _
from lino import DJANGO2
from etgen.html import E, tostring
from django.utils import translation
from django.utils import timezone
from django.conf import settings
from django.db import models
from datetime import timedelta
from datetime import datetime
from lxml import etree
from io import StringIO
import json
import logging

from django.core.serializers.json import DjangoJSONEncoder

logger = logging.getLogger(__name__)

html_parser = etree.HTMLParser()


def groupname(s):
    # Remove any invalid characters from the given string so that it can
    # be used as a Redis group name.
    # "Group name must be a valid unicode string containing only
    # alphanumerics, hyphens, or periods."

    s = s.replace('@', '-')
    return s.encode('ascii', 'ignore')


class ChatGroup(UserAuthored, Created, Referrable):
    class Meta(object):
        app_label = 'chat'
        verbose_name = _("Chat group")
        verbose_name_plural = _("Chat groups")
        ordering = ['created', 'id']

    title = dd.CharField(max_length=20)
    description = dd.RichTextField(max_length=200, blank=True, null=True);

    @dd.action(_("getChatGroups"))
    def getChatGroups(self, ar):
        """
        Returns info on all GroupChats for this user.
        """
        qs = ChatGroupMember.objects.filter(user=ar.get_user()).select_related("group")

        rows = [{"id": cp.group.pk,
                 "title": cp.group.title,
                 "unseen": cp.group.get_unseen_count(ar)} for cp in qs]

        return ar.success(rows=rows)

    def get_unseen_count(self, ar):
        """
        Returns count of messages that haven't been seen yet."""
        return ChatProps.objects.filter(chat__group=self, user=ar.get_user(), seen__isnull=True).count()

    @dd.action(_("Load GroupChat"))
    def loadGroupChat(self, ar):
        """Returns chat messages for a given chat"""
        rows = []
        for group in ar.selected_rows:
             last_ten = ChatProps.objects.filter(user=ar.get_user(),
                                                chat__group=self
                                                ).order_by(
                '-created').select_related("chat")
             rows.append({
                'title': group.title,
                'id': group.id,
                'messages': [cp.serialize(ar) for cp in last_ten]
            })
        return ar.success(rows=rows)


class ChatGroupMember(Created):
    class Meta(object):
        app_label = 'chat'

    user = dd.ForeignKey(
        'users.User',
        verbose_name=_("Author"),
        related_name="ChatGroups",
        blank=True, null=True)

    group = dd.ForeignKey(ChatGroup,

                          related_name="ChatGroupsMembers",
                          )


class ChatMessage(UserAuthored, Created, Previewable):
    class Meta(object):
        app_label = 'chat'
        verbose_name = _("Chat message")
        verbose_name_plural = _("Chat messages")
        ordering = ['created', 'id']

    # message_type = MessageTypes.field(default="change")

    seen = models.DateTimeField(_("seen"), null=True, editable=False)
    sent = models.DateTimeField(_("sent"), null=True, editable=False)
    group = dd.ForeignKey(
        'chat.ChatGroup', blank=True, null=True, verbose_name=_(u'Group'), related_name="messages")
    hash = dd.CharField(_("Hash"), max_length=25, null=True, blank=True)

    # body = dd.RichTextField(_("Body"), editable=False, format='html')

    def __str__(self):
        return "{}: {}".format(self.user, self.body)

        # return _("About {0}").format(self.owner)

    # return self.message
    # return _("Notify {0} about change on {1}").format(
    #     self.user, self.owner)

    def send_global_message(self):
        # message = {
        #     "id": self.pk,
        #     # "subject": str(self.subject),
        #     "user": self.user,
        #     "body": html2text(self.body),
        #     "created": self.created.strftime("%a %d %b %Y %H:%M"),
        # }
        logger.info("Sending Message %s:#%s" % (self.user, self.pk))
        send_global_chat(self)
        # self.sent = timezone.now()
        # self.save()

    @classmethod
    def markAsSeen(Cls, data):
        group_id = data['body'][0]
        # msg_ids = [chat[3] for chat in data['body'][1]]
        ChatProps.objects.filter(chat__group__pk=group_id, user=data['user'], seen__isnull=True).update(seen=timezone.now())


    @classmethod
    def onRecive(Cls, data):
        print(data)
        args = dict(
            user=data['user'],
            body=data['body']['body'],
            group_id=data['body']['group_id'],
            hash=data['body']['hash']
        )
        newMsg = Cls(**args)
        newMsg.full_clean()
        newMsg.save()
        newMsg.after_ui_create()
        newMsg.send_global_message()

    def after_ui_create(self, *args):
        for sub in self.group.ChatGroupsMembers.all().select_related("user"):
            props = ChatProps(chat=self, user=sub.user)
            props.full_clean()
            props.save()

        # super(ChatMessage, self).after_ui_save()

    @dd.action(_("ChatsMsg"))
    def getChats(self, ar):
        # return ar.success(rows=[])

        # doto, have work.
        last_ten_seen = ChatProps.objects.filter(seen__isnull=False).order_by('-created').select_related("chat")[:10]
        unseen = ChatProps.objects.filter(user=ar.get_user(), chat=self, seen__isnull=True).order_by(
            '-created').select_related("chat")

        last_ten_in_ascending_order = reversed(last_ten_seen)
        unseen_in_ascending_order = reversed(unseen)
        rows = [
            [(c.chat.user.username, ar.parse_memo(c.chat.body), c.chat.created, c.seen, c.chat.pk, c.chat.user.id) for c
             in last_ten_in_ascending_order],
            [(c.chat.user.username, ar.parse_memo(c.chat.body), c.chat.created, c.seen, c.chat.pk, c.chat.user.id) for c
             in unseen_in_ascending_order]
        ]
        return ar.success(rows=rows)


class ChatProps(UserAuthored, Created):
    class Meta(object):
        app_label = 'chat'
        # verbose_name = _("Chat message")
        # verbose_name_plural = _("Chat messages")
        ordering = ['user', 'chat']

    def serialize(self, ar=None):
        if ar is None:
            ar = BaseRequest()
        return (
        self.chat.user.username,
        ar.parse_memo(self.chat.body),
        json.loads(json.dumps(self.created, cls=DjangoJSONEncoder)),
        json.loads(json.dumps(self.seen, cls=DjangoJSONEncoder)),
        self.chat.pk,
        self.chat.user.id,
        self.chat.group.id,
        self.chat.hash)

    chat = dd.ForeignKey(ChatMessage, related_name="chatProps")
    seen = models.DateTimeField(_("seen"), null=True, editable=False, default=None)
    sent = models.DateTimeField(_("sent"), null=True, editable=False, default=None)



import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer

from lino.api import rt, dd

import logging
logger = logging.getLogger(__name__)


class ReactChatConsumer(WebsocketConsumer):
    def connect(self):
        self.accept()
        if self.scope.get('user', False):
            group_uid = str(self.scope["user"].pk) # ideally should be user.PK..
            self.add_group(group_uid)
            self.add_group("CHAT") # not used.


    def add_group(self, group_uid):
        """Connects this channel to a group
        Can add several groups"""
        async_to_sync(self.channel_layer.group_add)(group_uid, self.channel_name)

    def disconnect(self, close_code):
        pass

    def receive(self, text_data):
        try:
            data = json.loads(text_data)
            # logger.info("Recived WS data: " + text_data)

            user =self.scope.get('user', False)

            # Very Basic auth for confirming that this user is the on sending this message.
            if user:
                data["user"] = user
                if dd.is_installed("chat"):
                    ChatMessage = rt.models.resolve("chat.ChatMessage") # TODO HAVE WHITELIST OF FUNCTIONS!
                    if data.get('function') and hasattr(ChatMessage,data.get('function')):
                        getattr(ChatMessage,data.get('function'))(data)
        except Exception as E:
            logger.exception(E)
            raise E


    def send_notification(self, text):
        ## just passes data through. real work is done in .api
        self.send(text_data=text['text'])
